/* NSBitmapImageRep+XPM.m

   Methods for reading XPM images

   Copyright (C) 2006 Hubert Chan

   Written by:  Hubert Chan <hubert@uhoreg.ca>
   Date: Oct 2006

   This file is NOT (yet) part of the GNUstep GUI Library.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
   */

#include <Foundation/NSData.h>
#include <Foundation/NSArray.h>
#include "AppKit/NSImageRep.h"
#include "AppKit/NSColor.h"
#include "AppKit/NSGraphics.h"

#include "NSBitmapImageRep+XPM.h"

#include <string.h>
#include <stdlib.h>
#include <X11/xpm.h>

static NSString *XPMError(int errorCode)
{
  switch (errorCode)
  {
  case XpmColorError:
  case XpmColorFailed:
    return @"XPM color error";
  case XpmOpenFailed:
    return @"unable to open file";
  case XpmFileInvalid:
    return @"invalid XPM data";
  case XpmNoMemory:
    return @"out of memory";
  default:
    return @"unknown error code";
  }
}

typedef struct
{
  unsigned char data[4];
} color_t;

static void colorFromName(char *name, color_t *color, BOOL *hasAlpha)
{
  if (strcasecmp(name,"none") == 0)
  { /* transparent color */
    color->data[0] = color->data[1] = color->data[2] = 0;
    color->data[3] = 0;
    *hasAlpha = YES;
  }
  else
  { /* lookup names for X11 catalog */
    NSColor *col = [NSColor colorWithCatalogName: @"X11"
			      colorName: [NSString stringWithCString: name]];
    if (col == nil)
    { /* color not found: default to black */
      color->data[0] = color->data[1] = color->data[2] = 0;
      color->data[3]= 0xff;
    }
    else
    {
      color->data[0] = (char) ([col redComponent] * 0xff);
      color->data[1] = (char) ([col greenComponent] * 0xff);
      color->data[2] = (char) ([col blueComponent] * 0xff);
      color->data[3] = (char) ([col alphaComponent] * 0xff);
    }
  }
}

static void greyscaleColor(char *name, color_t *color, BOOL *hasAlpha)
{
  if (name[0] == '#')
  { /* hex code */
    char buf[3] = { 0,0,0 };
    char greyscale;
    int len = strlen(name) - 1;
    if (len == 0)
    { /* no color: default to black */
      color->data[0] = color->data[1] = color->data[2] = 0;
    }
    else if (len == 1)
    { /* one hex digit */
      buf[0] = name[1];
      greyscale = strtol(buf,NULL,16)*0x11;
      color->data[0] = color->data[1] = color->data[2] = greyscale;
    }
    else
    { /* two or more hex digits per component: only care about top byte */
      buf[0] = name[1];
      buf[1] = name[2];
      greyscale = strtol(buf,NULL,16)*0x11;
      color->data[0] = color->data[1] = color->data[2] = greyscale;
    }
    color->data[3] = 0xff; /* opaque */
  }
  else
  { /* named color */
    colorFromName(name,color,hasAlpha);
  }
}

@interface NSXPMImageRep (NSBitmapImageRep_XPM)
+ (BOOL) _bitmapIsXPM: (NSData *)imageData;
- (id) _initBitmapFromXPM: (NSData *)imageData
             errorMessage: (NSString **)errorMsg;
@end

@implementation NSXPMImageRep

+ (BOOL) _bitmapIsXPM: (NSData *)imageData
{
  char header[9];
  if ([imageData length] < 9)
    return NO;
  [imageData getBytes: header length: 9];
  switch (header[0])
  {
  case '!': /* possibly XPM version 2 */
    return strncmp(header+1," XPM2", 5) == 0 ? YES : NO;
  case '/': /* possibly XPM version 3 */
    return strncmp(header+1, "* XPM */", 8) == 0 ? YES : NO;
  }
  return NO;
}

#define SET_ERROR_MSG(msg) \
   if (errorMsg != NULL) \
     {\
       *errorMsg = msg; \
     }\
   NSLog(msg);

#define CREATE_ERROR_MSG(function, failedop, msg) \
   { \
     NSString *err = [NSString stringWithFormat: @"%@: failed to %@: %@", \
			       function, failedop, msg]; \
     SET_ERROR_MSG(err); \
     if (loaded) \
     { \
       XpmFreeXpmImage(&image); \
     } \
     if (colormap) \
       free(colormap); \
     RELEASE(self); \
     return nil; \
   }

- (id) _initBitmapFromXPM: (NSData *)imageData
             errorMessage: (NSString **)errorMsg
{
#define _FUNC @"_initBitmapFromXPM:errorMessage"
  const char *buffer = [imageData bytes];
  XpmImage image;
  int rc;
  BOOL loaded = NO;
  color_t *colormap = NULL;
  int i;
  BOOL hasAlpha = NO;
  id rv;

  if (buffer == NULL)
    CREATE_ERROR_MSG(_FUNC, @"load image", @"no image data");

  if ((rc = XpmCreateXpmImageFromBuffer ((char *)buffer, &image, NULL))
      != XpmSuccess)
    CREATE_ERROR_MSG(_FUNC,@"create XpmImage", XPMError(rc));
  loaded = YES;

  colormap = (color_t *)malloc(image.ncolors * sizeof(color_t));

  for (i = 0; i < image.ncolors; i++)
  {
    XpmColor *color = image.colorTable + i;
    if (color->c_color)
    {
      if (color->c_color[0] == '#')
      { /* hex code */
	char buf[3] = { 0,0,0 };
	int len = (strlen(color->c_color) - 1) / 3; /* length of each component */
	if (len == 0)
	{ /* no color: default to black */
	  colormap[i].data[0] = colormap[i].data[1] = colormap[i].data[2] = 0;
	}
	else if (len == 1)
	{ /* one hex digit per component */
	  int j;
	  for (j = 0; j < 3; j++)
	  {
	    buf[0] = color->c_color[1+j];
	    colormap[i].data[j] = strtol(buf,NULL,16)*0x11;
	  }
	}
	else
	{ /* two or more hex digits per component: only care about top byte */
	  int j;
	  for (j = 0; j < 3; j++)
	  {
	    buf[0] = color->c_color[1+j*len];
	    buf[1] = color->c_color[2+j*len];
	    colormap[i].data[j] = strtol(buf,NULL,16);
	  }
	}
	colormap[i].data[3] = 0xff; /* opaque */
      }
      else
      { /* named color */
	colorFromName(color->c_color,colormap+i,&hasAlpha);
      }
    }
    else if (color->g_color)
    {
      greyscaleColor(color->g_color,colormap+i,&hasAlpha);
    }
    else if (color->g4_color)
    {
      greyscaleColor(color->g_color,colormap+i,&hasAlpha);
    }
    else if (color->m_color)
    {
      greyscaleColor(color->g_color,colormap+i,&hasAlpha);
    }
    else
    { /* no color: default to black */
      colormap[i].data[0] = colormap[i].data[1] = colormap[i].data[2] = 0;
	colormap[i].data[3] = 0xff; /* opaque */
    }
  }

  /* copy bitmap data */
  if (hasAlpha)
  {
    int x,y,c;
    unsigned char *dest;
    unsigned int *src = image.data;
    rv = [self initWithBitmapDataPlanes: NULL
	       pixelsWide: image.width pixelsHigh: image.height
	       bitsPerSample: 8 samplesPerPixel: 4
	       hasAlpha: hasAlpha isPlanar: NO
	       colorSpaceName: NSCalibratedRGBColorSpace
	       bytesPerRow: image.width * 4
	       bitsPerPixel: 8 * 4];

    dest = [rv bitmapData];
    for (y = 0; y < image.height; y++)
    {
      for (x = 0; x < image.width; x++)
      {
	unsigned int color = *(src++);
	for (c = 0; c < 4; c++)
	{
	  *(dest++) = colormap[color].data[c];
	}
      }
    }
  }
  else
  {
    int x,y,c;
    unsigned char *dest;
    unsigned int *src = image.data;
    rv = [self initWithBitmapDataPlanes: NULL
	       pixelsWide: image.width pixelsHigh: image.height
	       bitsPerSample: 8 samplesPerPixel: 3
	       hasAlpha: hasAlpha isPlanar: NO
	       colorSpaceName: NSCalibratedRGBColorSpace
	       bytesPerRow: image.width * 3
	       bitsPerPixel: 8 * 3];

    dest = [rv bitmapData];
    for (y = 0; y < image.height; y++)
    {
      for (x = 0; x < image.width; x++)
      {
	unsigned int color = *(src++);
	for (c = 0; c < 3; c++)
	{
	  *(dest++) = colormap[color].data[c];
	}
      }
    }
  }

  XpmFreeXpmImage(&image);
  free(colormap);

  return rv;

#undef _FUNC
}

+ (void) load
{
  [NSImageRep registerImageRepClass: [NSXPMImageRep class]];
}

+ (BOOL) canInitWithData: (NSData *) imageData
{
  if (imageData == nil)
    return NO;
  return [self _bitmapIsXPM: imageData];
}

+ (NSArray *) imageUnfilteredFileTypes
{
  static NSArray *types = nil;

  if (types == nil)
  {
    types = [[NSArray alloc] initWithObjects: @"xpm", nil];
  }

  return types;
}

+ (NSArray *) imageRepsWithData: (NSData *) imageData
{
  if (imageData == nil)
    {
      NSLog(@"NSBitmapImageRep: nil image data");
      return [NSArray array];
    }

  if ([self _bitmapIsXPM: imageData])
  {
    NSXPMImageRep *rep;
    NSArray *a;

    rep = [[self alloc] _initBitmapFromXPM: imageData
			errorMessage: NULL];

    if (rep == nil)
      return [NSArray array];
    a = [NSArray arrayWithObject: rep];
    DESTROY(rep);
    return a;
  }
  return [NSArray array];
}

- (id) imageWithData: (NSData *) imageData
{
  if (imageData == nil)
  {
    RELEASE(self);
    return nil;
  }

  if ([isa _bitmapIsXPM: imageData])
    return [self _initBitmapFromXPM: imageData
		 errorMessage: NULL];

  RELEASE(self);
  return nil;
}

@end
